#include <LiquidCrystal.h>
#include <PID_v1.h> // https://github.com/br3ttb/Arduino-PID-Library/

#define STEPPER_DIR     8
#define STEPPER_STEP    9

#define RIGHTBUTTON     6
#define LEFTBUTTON      4
#define STARTBUTTON     2

#define ECHOPIN         30
#define TRIGPIN         32
#define DELAY           500

#define POTKP           A0
#define POTKI           A1
#define POTKD           A2

#define LCD_RS          47
#define LCD_ENABLE      46
#define LCD_D4          36
#define LCD_D5          37
#define LCD_D6          38
#define LCD_D7          39

LiquidCrystal lcd(LCD_RS, LCD_ENABLE, LCD_D4, LCD_D5, LCD_D6, LCD_D7);

struct Status 
{
    int state;
    int position;
    int setpoint;
    int Kp;
    int Ki;
    int Kd;
};

void display(struct Status *status)
{
    lcd.clear();

    lcd.setCursor(0, 0),
    lcd.print("Kp: ");

    lcd.setCursor(4, 0);
    lcd.print(status->Kp);

    lcd.setCursor(8, 0);
    lcd.print("Ki: ");

    lcd.setCursor(12, 0);
    lcd.print(status->Ki);

    lcd.setCursor(0, 1);
    lcd.print("Kd: ");

    lcd.setCursor(4, 1);
    lcd.print(status->Kd);

    lcd.setCursor(8, 1);
    lcd.print("M: ");

    lcd.setCursor(11, 1);

    if (status->state == 2)
    {
        lcd.print("PID");
    }

    else if (status->state == 1)
    {
        lcd.print("TUNE");    
    }

    else
    {
        lcd.print("ZERO");
    }

    delay(10);
}

double ping(void)
{
    double distance;

    digitalWrite(TRIGPIN, LOW);
    delayMicroseconds(2);

    digitalWrite(TRIGPIN, HIGH);
    delayMicroseconds(2);

    digitalWrite(TRIGPIN, LOW);
    distance = pulseIn(ECHOPIN, HIGH, 6000) / 58.2;

    if (distance == 0 || distance > 45)
    {
        ping();
    }

    else
    {
        return distance;
    }
}

void readKValues(struct Status *status)
{
    if (status->state == 0 || status->state == 1)
    {
        status->Kp = analogRead(POTKP);
        status->Ki = analogRead(POTKI);
        status->Kd = analogRead(POTKD);

        status->Kp = map(status->Kp, 0, 1023, 0, 300);
        status->Ki = map(status->Ki, 0, 1023, 0, 300);
        status->Kd = map(status->Kd, 0, 1023, 0, 300);

        display(status);
    }

    else
    {
        display(status);
    }
}

void rightButton(struct Status *status)
{
    if (status->state >= 1)
    {
        if (status->position <= 150)
        {
            turnRight();
            status->position++;
        }

        else
        {
            stop();
        }
    }

    else
    {
        turnRight();
    }
}

void leftButton(struct Status *status)
{
    if (status->state >= 1)
    {
        if (status->position >= -150)
        {
            turnLeft();
            status->position--;
        }

        else
        {
            stop();
        }
    }

    else
    {
        turnLeft();
    }
}

void PIDMode(struct Status *status)
{
    double input;
    double output;
    double setpoint;

    double Kp;
    double Ki;
    double Kd;

    status->position = 0;
    Kp = (double) status->Kp;
    Ki = (double) status->Ki;
    Kd = (double) status->Kd;
    setpoint = (double) status->setpoint;  

    PID myPID(&input, &output, &setpoint, Kp, Ki, Kd, DIRECT);
    myPID.SetMode(AUTOMATIC);

    for(;;)
    {
        if (digitalRead(STARTBUTTON) == HIGH)
        {
            status->state = 0;
            break;
        }

        else
        {
            input = ping();
            myPID.Compute();

            if (output >= 127)
            {
                if (status->position <= 100)
                {
                    turnRight();
                    status->position++;
                }
            }

            else if (output < 127)
            {
                if (status->position >= -100)
                {
                    turnLeft();
                    status->position--;
                }
            }
        }
    }
}

void control(struct Status *status)
{
    readKValues(status);

    if (status->state == 2)
    {
        PIDMode(status);
    }

    if (digitalRead(STARTBUTTON) == HIGH)
    {
        if (status->state == 0)
        {
            status->state = 1;
            status->position = 0;
            delay(500);
        }

        else if (status->state == 1)
        {
            status->state = 2;
            status->position = 0;
            delay(500);
        }
    }

    if (digitalRead(RIGHTBUTTON) == HIGH)
    {
        rightButton(status);
    }

    if (digitalRead(LEFTBUTTON) == HIGH)
    {
        leftButton(status);
    }

    else
    {
        stop();
    }
}

void turnRight(void)
{
    digitalWrite(STEPPER_DIR, HIGH);
    digitalWrite(STEPPER_STEP, HIGH);
    delayMicroseconds(DELAY);
    digitalWrite(STEPPER_STEP, LOW);
    delayMicroseconds(DELAY);
}

void turnLeft(void)
{
    digitalWrite(STEPPER_DIR, LOW);
    digitalWrite(STEPPER_STEP, HIGH);
    delayMicroseconds(DELAY);
    digitalWrite(STEPPER_STEP, LOW);
    delayMicroseconds(DELAY);
}

void stop(void)
{
    digitalWrite(STEPPER_DIR, LOW);
    delayMicroseconds(500);
    digitalWrite(STEPPER_STEP, LOW);
    delayMicroseconds(500);    
}

int main(void)
{
    init();

    pinMode(STEPPER_DIR, OUTPUT); 
    pinMode(STEPPER_STEP, OUTPUT);
    
    digitalWrite(STEPPER_DIR, LOW);
    digitalWrite(STEPPER_STEP, LOW);

    pinMode(RIGHTBUTTON, INPUT);
    pinMode(LEFTBUTTON, INPUT);
    pinMode(STARTBUTTON, INPUT);

    pinMode(TRIGPIN, OUTPUT);
    pinMode(ECHOPIN, INPUT);

    lcd.begin(16, 2);
    lcd.clear();
    lcd.noCursor();
    lcd.noBlink();
    lcd.noAutoscroll();

    // state 0 = free adjust, state 1 = locked adjust, state 2 = running.
    struct Status status;
    status.state = 0;
    status.position = 0;
    status.setpoint = 21;

    for(;;)
    {
        control(&status);
    }

    return 0;
}
